# `api` Mod

This mod implements all of the RaspberryJuice extensions to the MCPI API, for the modding API, see [misc/api.cpp](../misc/api.cpp).

This includes:
- [x] `world.getBlocks(x0: int, y0: int, z0: int, x1: int, y1: int, z1: int) -> int[]`
  - Gets all the block ids between x0, y0, z0 and x1, y1, z1
- [x] `world.getPlayerId(name: str) -> int`
  - Gets the id of the first player who has the same name as `name`. "Fail" is returned on failure.
- [x] `entity.getName(id: int) -> str`
  - Gets the name of the entity of `id`, as MCPI does not have name tags, it will just be the entity type for non-players
- [x] `world.getEntities(type: int) -> {id: int, type: int, name: str, x: float, y: float, z: float}[]`
  - Returns a list of entities
- [x] `world.removeEntity(id: int) -> int`
  - Removes a single entity and returns 1 if successful
- [x] `world.removeEntities(type: int) -> int`
  - Removes all entities of a type, and returns how many were removed
- [x] `events.chat.posts() -> {"0", message}[]`
  - Gets a list of chat messages, see the limitations section down below to see why it has a zero
  - When compatibility mode is enabled, `|` in the chat messages is replaced with `\`
  - When compatibility mode is disabled, `|` in the chat messages is replaced with `\|`
- [x] `player.events.chat.posts() -> {"0", message}[]`
 - This is *exactly* the same as `events.chat.posts`, however it only returns chat messages sent by the player
- [x] `events.projectile.hits() -> {x: int, y: int, z: int, "1", owner: string, target: int}` (and `player.events.projectile.hits()` and `entity.events.projectile.hits(id: int)`)
  - Returns a list of projectile hit events, the list is cleared each time it is read by this call
  - When target id is not 0, it means the projectile hit an entity and (x, y, z) is the entity's position
  - When target id is 0, it means the projectile hit a block and (x, y, z) is the block's position
  - When compatibility mode is disabled, the constant "1" is removed in the return value, leaving just the useful information
  - When compatibility mode is disabled, the owner's name is the owner's id instead, as players can have duplicate names
  - Unlike RaspberryJuice, the last argument is the hit entity id, not the entity's name. This shouldn't break anything, and indeed should be more useful.
- [x] `player.setDirection(x: float, y: float, z: float)`
  - sets rotation as if the player was at 0, 0, 0 and looking towards (x, y, z)
- [x] `entity.setDirection(id: int, x: float, y: float, z: float)`
  - Sets rotation as if the entity of `id` was at 0, 0, 0 and looking towards (x, y, z)
- [x] `player.getDirection() -> {x: float, y: float, z: float}`
  - Gets the location the player would be looking at if they were at 0, 0, 0
- [x] `entity.getDirection() -> {x: float, y: float, z: float}`
  - Gets the location the entity of `id` would be looking at if they were at 0, 0, 0
- [x] `player.setRotation(id: int)`
  - Sets the yaw of the player
- [x] `entity.setRotation(id: int)`
  - Sets the yaw of the entity of `id`
- [x] `player.getRotation`
  - Gets the yaw of the player
- [x] `entity.getRotation`
  - Gets the yaw of the entity of `id`
- [x] `player.getPitch`
  - Gets the pitch of the player
- [x] `entity.getPitch`
  - Gets the pitch of the entity of `id`
- [x] `player.setPitch(id: int)`
  - Sets the pitch of the player
- [x] `entity.setPitch(id: int)`
  - Sets the pitch of the entity of `id`
- [x] `player.getEntities(dist: int, type: int) -> {id: int, type: int, name: str, x: float, y: float, z: float}[]`
  - Gets all entities within `dist` of the player if the entity is of `type` (if `type` is -1, all entity types are included).
  - This can include the player.
- [x] `entity.getEntities(id: int, dist: int, type: int) -> {id: int, type: int, name: str, x: float, y: float, z: float}[]`
  - Gets all entities within `dist` of the entity of `id` if the entity is of `type` (if `type` is -1, all entity types are included).
  - This can include the entity.
- [x] `player.removeEntities(dist: int, type: int) -> int`
  - Removes all entities within `dist` of the player if the entity is of `type` (if `type` is -1, all entity types are included).
  - This will not include any players.
  - It returns the number of entities removed.
- [x] `entity.removeEntities(id: int, dist: int, type: int) -> int`
  - Removes all entities within `dist` of the entity of `id` if the entity is of `type` (if `type` is -1, all entity types are included).
  - This will not include any players, but may include the entity calling it if the entity is not a player.
  - It returns the number of entities removed.
- [x] `world.setSign(x: int, y: int, z: int, id: int, data: int, [l1: str], [l2: str], [l3: str], [l4: str])`
  - Sets a block of `id:data` at the specified point, if the block is a sign, it will attempt to set lines 1 through 4 of the sign to the given text
  - For the API, the lines must be below 100 characters, however when loading signs MCPI will cap it at 16 characters (this can be disabled by patching out the call at `0xd1e2c`).
  - The lines are optional
  - The wiki has a [list of blocks](https://mcpirevival.miraheze.org/wiki/Minecraft:_Pi_Edition_Complete_Block_List), sign is 63 and wall sign is 68
- [x] `world.spawnEntity(x: int, y: int, z: int, type: int) -> int`
  - Spawns an entity of `type` at the given position
  - Entities with a type of 0 cannot be spawned
  - The list of entity types for RaspberryJuice can be found [here](https://github.com/zhuowei/RaspberryJuice/blob/master/src/main/resources/mcpi/api/python/modded/mcpi/entity.py). This list is used when compatibility mode is enabled.
  - The list of entity types can be found by running the command below, or on the [wiki](https://mcpirevival.miraheze.org/wiki/Minecraft:_Pi_Edition_Complete_Entity_List). This list is used when compatibility mode is disabled.
- [x] `world.getEntityTypes() -> {type: int, name: str}`
  - Returns a list of known entity types, if there are modded entities this list may be incorrect
  - The ids return depend on if compatibility mode is enabled, see `world.spawnEntity` for more
- [x] `player.getAbsPos() -> {x: int, y: int, z: int}`
  - Gets the absolute position of the player, aka, without passing it through `OffsetPosTranslator`
- [x] `player.setAbsPos(x: int, y: int, z: int)`
  - Sets the absolute position of the player, aka, without passing it through `OffsetPosTranslator`
- [x] `player.events.block.hits()` and `entity.events.block.hits(id: int)`, both return `{x: int, y: int, z: int, face: int, hitter: int}`
  - Same as `events.block.hits`, but for only the player or entity with the id. Only block hits in creative with an iron sword are counted
- [x] `player.events.clear()` and `entity.events.clear(id: int)`
  - These clear projectile, block hit, and chat events for the player or entity
- [x] `events.clear` has been changed
  - It will clear block hit events and chat events, and when compat mode is disabled; projectile events

Due to MCPI limitations:
- `events.chat.posts`'s id field is set to 0 for messages that did not come from the main player
- `events.block.hits` (and more specialized calls), only work in creative or creator mode

Egdecases:
- `world.removeEntities`/`player.removeEntities`/`entity.removeEntities` will not remove players when given 0 or -1 for type.
- `entity.getName` will not get the name of non-player entities with a type id of 0 due to ambiguity.
- All Raspberry Juice commands/responses involving the player name are designed around the MCJE username restrictions, not the much looser MCPI restrictions. They may cause problems.
- Chat messages will have `|` turned into `\` to prevent bad parsing


Extras:
- Compatibility mode, this is enabled by default, and can be changed with `reborn.disableCompatMode` or `reborn.enableCompatMode`.
  - `events.clear` 

## How does it work?

![Reborn-API.png](Reborn-API.png)
